/****************************************************************************
**
** Copyright (C) 2024 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Design Viewer of the Qt Toolkit.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
****************************************************************************/

#include "qrscanner.h"

#include <QApplication>
#include <QCamera>
#include <QCameraDevice>
#include <QElapsedTimer>
#include <QImageCapture>
#include <QMediaDevices>
#include <QMessageBox>
#include <QPermission>
#include <QVideoSink>
#include <QtConcurrent>

#include "../3rdparty/zxing-cpp/example/ZXingQtReader.h"

using namespace ZXingQt;

QrScanner::~QrScanner()
{
    if (m_captureSession) {
        m_captureSession->camera()->stop();
        m_captureSession->camera()->deleteLater();
        m_captureSession->videoOutput()->deleteLater();
        m_captureSession.clear();
    }
}

void QrScanner::scanQrCode()
{
    // request permissions
    QCameraPermission permission;
    QCoreApplication &app = *QCoreApplication::instance();
    switch (app.checkPermission(permission)) {
    case Qt::PermissionStatus::Granted:
        openCamera();
        break;
    case Qt::PermissionStatus::Undetermined:
    case Qt::PermissionStatus::Denied:
        app.requestPermission(permission, [this](const QPermission &permission) {
            if (permission.status() == Qt::PermissionStatus::Denied) {
                QMessageBox msgBox{QMessageBox::Critical,
                                   "Critical:",
                                   "Camera permission denied",
                                   QMessageBox::Ok};
                msgBox.exec();
                return;
            }

            openCamera();
        });
        break;
    }
}

void QrScanner::openCamera()
{
    // start camera
    m_captureSession.reset(new QMediaCaptureSession(this));
    m_captureSession->setCamera(new QCamera(this));
    m_captureSession->setVideoOutput(new CustomVideoWidget);

    m_captureSession->camera()->setFocusMode(QCamera::FocusModeAuto);

    CustomVideoWidget *videoWidget = qobject_cast<CustomVideoWidget *>(
        m_captureSession->videoOutput());

    connect(m_captureSession->videoSink(),
            &QVideoSink::videoFrameChanged,
            this,
            [&](const QVideoFrame &frame) {
                static int i = 0;
                static QAtomicInt running = 0;

                if (i++ < 20 || running > 0) {
                    return;
                }

                i = 0;

                QtConcurrent::run([=] {
                    // lock the thread so that only one barcode can be read at a time
                    running++;

                    QElapsedTimer timer;
                    timer.start();
                    QList<Result> results = ReadBarcodes(frame.toImage());
                    qDebug() << "Barcode detection took" << timer.elapsed() << "ms";

                    // release the lock so that another barcode can be read
                    running--;
                    return results;
                }).then([=](const QList<Result> &results) {
                    if (results.isEmpty() || !m_captureSession)
                        return;

                    qDebug() << "Stopping camera";
                    qobject_cast<CustomVideoWidget *>(m_captureSession->videoOutput())->close();
                    qDebug() << "Camera stopped";

                    Result result = results.first();

                    qDebug() << "Text:   " << result.text();
                    qDebug() << "Format: " << result.format();
                    qDebug() << "Content:" << result.contentType();

                    // we have to use invokeMethod because we are in a different thread
                    // then where the serviceConnector is created
                    // QMetaObject::invokeMethod(this,
                    //                           "parseDesignViewerUrl",
                    //                           Qt::QueuedConnection,
                    //                           Q_ARG(QUrl, result.text()));
                    emit qrCodeScanned(result.text());
                });
            });

    // stop camera when app is not active.
    // i.e. when the user switches to another app or the screen is locked
    QGuiApplication *app(qobject_cast<QGuiApplication *>(QCoreApplication::instance()));
    connect(app, &QGuiApplication::applicationStateChanged, this, [&](Qt::ApplicationState state) {
        if (state != Qt::ApplicationState::ApplicationActive && m_captureSession) {
            qDebug() << "Application is not active. Stopping camera";
            qobject_cast<CustomVideoWidget *>(m_captureSession->videoOutput())->close();
        }
    });

    // stop camera when video widget is closed
    connect(videoWidget, &CustomVideoWidget::closed, this, [&] {
        qDebug() << "Video widget closed. Clearing capture session";
        m_captureSession->camera()->stop();
        m_captureSession->camera()->deleteLater();
        m_captureSession->videoOutput()->deleteLater();
        m_captureSession.clear();
    });

    videoWidget->setWindowFlags(Qt::Window | Qt::WindowStaysOnTopHint);
    videoWidget->show();
    m_captureSession->camera()->start();
}
